from collections import deque
import doctest
import os
from types import ModuleType
from unittest import TestCase

import CommonMark

import seen
Seen = seen.Seen


EXAMPLE_LENGTH = 4
KEY_FUNCTIONS = (hash, id, lambda: None)
OBJECTS_INTEGERS = list(range(EXAMPLE_LENGTH))
OBJECTS_LISTS = [[x] for x in range(EXAMPLE_LENGTH)]
EXAMPLES = (
    (hash, OBJECTS_INTEGERS),
    (id, OBJECTS_LISTS),
)


def iter_markdown_block(b):
    p = deque([b])
    while p:
        b = p.popleft()
        yield b
        p.extend(b.children)


def doctest_string(s):
    m = ModuleType('doctest_string')
    m.__doc__ = s
    return doctest.testmod(m)


class TestSeen(TestCase):
    def test_instantiation_key_provided(self):
        for key_func in KEY_FUNCTIONS:
            self.assertIsInstance(Seen(key_func), Seen)

    def test_instantiation_key_omitted(self):
        with self.assertRaises(TypeError):
            Seen()

    def test_membership_empty(self):
        for key_func, data in EXAMPLES:
            for obj in data:
                s = Seen(key_func)
                self.assertFalse(obj in s)

    def test_membership_rolling(self):
        for key_func, data in EXAMPLES:
            s = Seen(key_func)
            for obj in data:
                self.assertFalse(obj in s)
                self.assertTrue(obj in s)

    def test_len(self):
        for key_func, data in EXAMPLES:
            s = Seen(key_func)
            for i, obj in enumerate(data):
                self.assertEqual(len(s), i)
                obj in s
                self.assertEqual(len(s), i + 1)

    def test_clear(self):
        for key_func, data in EXAMPLES:
            for obj in data:
                s = Seen(key_func)
                obj in s
                s.clear()
                self.assertEqual(len(s), 0)
                self.assertFalse(obj in s)

    def test_repr(self):
        for key_func, data in EXAMPLES:
            s = Seen(key_func)
            for i, obj in enumerate(data):
                self.assertEqual(repr(s), '<Seen ({})>'.format(i))
                obj in s
                self.assertEqual(repr(s), '<Seen ({})>'.format(i + 1))

    def test_doctest(self):
        r = doctest.testmod(seen)
        self.assertEqual(r.failed, 0)

    def test_readme(self):
        readme = os.path.join(os.path.dirname(seen.__file__), 'README.md')
        parser = CommonMark.DocParser()
        with open(readme) as f:
            md_ast = parser.parse(f.read())
        for b in iter_markdown_block(md_ast):
            if b.t == 'FencedCode' and b.info == 'python':
                r = doctest_string(b.string_content)
                self.assertEqual(r.failed, 0)
